;;; -*- Mode: Common-Lisp; Base:10; Package: http -*-

;;; File: http.cl
;;; Program: 

;;; Author: Alberto Riva (alb@aim.unipv.it)
;;; Created: Tue May  9, 1995
;;; Updated: Wed Nov 15, 1995

;;; Contents:
;;; -*- -*-

(in-package :http)

(defun init-streams ()
  (close-streams)
  (if (stringp *LispWeb-log-file*)
      (setq *LispWeb-log-stream* (open (merge-pathnames *LispWeb-log-file* *LispWeb-conf-dir*)
				       :direction :output
				       :if-does-not-exist :create
				       :if-exists :append))
    (setq *LispWeb-log-stream* *LispWeb-log-file*))
  (if (stringp *LispWeb-debug-file*)
      (setq *LispWeb-debug-stream* (open (merge-pathnames *LispWeb-debug-file* *LispWeb-conf-dir*)
					 :direction :output
					 :if-does-not-exist :create
					 :if-exists :append))
    (setq *LispWeb-debug-stream* *LispWeb-debug-file*))
  (if (stringp *LispWeb-timing-file*)
      (setq *LispWeb-timing-stream* (open (merge-pathnames *LispWeb-timing-file* *LispWeb-conf-dir*) 
					  :direction :output
					  :if-does-not-exist :create
					  :if-exists :append))
    (setq *LispWeb-timing-stream* *LispWeb-timing-file*)))

(defun init-logger ()
  (lw-set-log :http *LispWeb-log-file*)
  (lw-set-log :debug *LispWeb-debug-file*)
  (lw-set-log :time *LispWeb-timing-file*))

(defun close-streams ()
  "Close all the logging streams."
  (maybe-close-stream *LispWeb-log-stream*)
  (maybe-close-stream *LispWeb-debug-stream*)
  (maybe-close-stream *LispWeb-timing-stream*))

(defun maybe-close-stream (stream)
  "Close STREAM only if it is associated with a file
  (i.e., not T, NIL, or a *terminal-io*)."
  (when (and (streamp stream)
	     (not (eq stream *terminal-io*)))
    (close stream)))

;;; Logging


(defgeneric http-log (object)
  (:documentation "Write a line relative to the current request in the log file."))
   
(defmethod http-log (unknown)
  (declare (ignore unknown))
  (lw-log :http-debug "Connection closed prematurely."))

(defmethod http-log ((message string))
  "Writes a string to the log stream."
  (lw-log :http "~a - - [~a ~a] ~s [~d ~d]~%"
	  (get-host-by-addr (http-info-addr *http-info*))
	  (short-date) 
	  (short-time)
	  message
	  (http-info-request-counter *http-info*)
	  (http-info-code *http-info*)))

(defmethod http-log ((info http-info))
  "Writes the request to the log stream."
  (http-log (http-info-request-line info)))

(defun http-log-message (&rest log-message)
  "Send a message to the logfile."
  (http-log (apply #'format (cons nil log-message))))

(defun http-debug-message (&rest debug-message)
  "Send a string to the debug file."
  (lw-log :debug "[~a ~a] ~s~%"
	  (short-date)
	  (short-time) 
	  (apply #'format (cons nil debug-message))))

(defun http-timing-message (request time)
  "Send a string to the timing file."
  (lw-log :time "[~a ~a] ~s ~a~%" (short-date) (short-time) request time))

(defun lw-debug (&optional mode)
  "Set debugging to MODE, that can be :on or :off. With no argument,
toggles the debug state. When debugging is :on, the Lisp debugger will 
be invoked in case of errors."
  (case mode 
    (:on (setf *lispweb-debug* t))
    (:off (setf *lispweb-debug* nil))
    (t (setf *lispweb-debug* (not *lispweb-debug*)))))

(defun lw-trace (&rest pages)
  "Add PAGES to the list of traced pages. When a traced page is called,
debugging information will be printed to the server's debug stream."
  (mapcar #'(lambda (name)
	      (pushnew name *LispWeb-traced-pages*))
	  pages)
  *LispWeb-traced-pages*)

(defun lw-untrace (&rest pages)
  "Remove PAGES from the list of traced pages."
  (mapcar #'(lambda (name)
	      (setq *LispWeb-traced-pages*
		(delete name *LispWeb-traced-pages*)))
	  pages)
  *LispWeb-traced-pages*)

(defun page-traced-p (page)
  (member page *LispWeb-traced-pages*))

(defun print-page-trace (name)
  (http-debug-message "~a called from URL: ~s" name (http-info-request-object *http-info*))
  (http-debug-message "~a arguments: ~s" name (http-info-arguments *http-info*))
  (when (http-info-fields *http-info*)
    (do ((h (http-info-fields *http-info*) (cddr h)))
	((null h))
      (http-debug-message "~a field ~:(~a~): ~s" name (car h) (second h)))))

(defun send-error-message (stream condition &rest args)
  (format stream "Error: ~a ~{~a ~}~%" condition args))