;;; -*- Mode: Common-Lisp; Base:10; Package: http -*-

;;; File: http.cl
;;; Program: 

;;; Author: Alberto Riva (alb@aim.unipv.it)
;;; Created: Tue May  9, 1995
;;; Updated: Wed Nov 15, 1995

;;; Thanks to: Mark Kantrowitz (mkant@cs.cmu.edu), 
;;;            Joe Konstan (konstan@cs.berkeley.edu), 
;;;            Giordano Lanzola (giordano@aim.unipv.it)
;;;            John Mallery (jcma@ai.mit.edu)

;;; Updated: Fri Apr 11, 1997, by Alberto Riva (alb@aim)
;;; Contents:
;;;   Top-level     IN-PACKAGE 
;;;   Function      INIT-SERVER 
;;;   Function      LOAD-INIT-FILE 
;;;   Function      LOAD-PERSONAL-INIT-FILE 
;;;   Function      INIT-STREAMS 
;;;   Function      INIT-LOGGER 
;;;   Function      CLOSE-STREAMS 
;;;   Function      MAYBE-CLOSE-STREAM 
;;;   Function      START-TCP-SERVICES 
;;;   Generic       HTTP-LOG 
;;;   Method        HTTP-LOG 
;;;   Method        HTTP-LOG 
;;;   Function      HTTP-LOG-MESSAGE 
;;;   Function      HTTP-DEBUG-MESSAGE 
;;;   Function      HTTP-TIMING-MESSAGE 
;;;   Function      LISPWEB-PATH 
;;;   Function      ENCRYPT-PASSWORD 
;;;   Function      CHECK-PASSWORD 
;;;   Function      HTTP-TIME 
;;;   Function      SHORT-DATE 
;;;   Function      SHORT-TIME 
;;;   Function      SPLICE-STRING 
;;;   Function      SAFE-READNUM 
;;;   Function      HTTP-TO-UNIVERSAL-TIME 
;;;   Function      HTTP-READ-LINE 
;;;   Function      PROCESS-REQUEST 
;;;   Function      LW-DEBUG 
;;;   Function      LW-TRACE 
;;;   Function      LW-UNTRACE 
;;;   Function      PAGE-TRACED-P 
;;;   Function      PRINT-PAGE-TRACE 
;;;   Function      PROCESS-HEADERS 
;;;   Function      MAKE-OPTION-PAIR 
;;;   Function      PARSE-HEADER-OPTIONS 
;;;   Function      FIND-HEADER 
;;;   Function      HEADER-VALUES 
;;;   Function      HEADER-VALUE 
;;;   Function      ADD-HEADER-VALUE 
;;;   Function      SET-HEADER-VALUES 
;;;   Function      PRINT-HEADER-FIELD 
;;;   Function      PROCESS-FORM-FIELDS 
;;;   Function      PROCESS-FORM-FIELDS-AUX 
;;;   Function      FORM-VALUE 
;;;   Function      FORM-VALUES 
;;;   Function      ADD-FORM-VALUE 
;;;   Function      SET-FORM-VALUES 
;;;   Function      EXPLODE-REQUEST 
;;;   Function      TRANSLATE-REQUEST 
;;;   Condition     NO-AUTHORIZATION 
;;;   Condition     BAD-ACCESS-KEY 
;;;   Condition     ACCESS-KEY-EXPIRED 
;;;   Variable      *NO-AUTH-CONDITION* 
;;;   Variable      *BAD-KEY-CONDITION* 
;;;   Variable      *KEY-EXPIRED-CONDITION* 
;;;   Function      MAKE-AUTH-LIST 
;;;   Function      MATCH-AUTH 
;;;   Function      TEST-AUTH 
;;;   Function      CHECK-AUTH 
;;;   Function      TEST-ACCESS-KEY 
;;;   Function      HTTP-CLIENT 
;;; -*- -*-

(in-package :http)

(defun init-server (&optional options)
  "Load init file, set system variables according to command-line
  arguments, encrypt the password and initialize the server streams. 
The initialization process takes place in the following way:
1) Determine the name and directory of the initialization file from
   the -init and -cdir command-line arguments;
2) Load the global initialization file;
3) Load the personal initialization file;
4) Parse the remaining command-line arguments."

  (let ((conf-dir (second (member '*LispWeb-conf-dir* options))) ;-cdir option supplied?
	(init-file (second (member '*LispWeb-init-file* options)))) ;-i option supplied?
    (when init-file (setq *LispWeb-init-file* init-file)) ;override default
    (when conf-dir (setq *LispWeb-conf-dir* conf-dir)))

  (load-init-file)
  (load-personal-init-file)
  
  (let ((to-set nil))			;set variables
    (mapc #'(lambda (opt)
	      (cond 
	       (to-set (set to-set opt)
		       (setq to-set nil))
	       (t (setq to-set opt))))
	  options))
  (setq *operator-password* (encrypt-password *operator-password*))
  (setq *startup-date* (http-time))
  (init-logger))

(defun load-init-file ()
  (let ((path (merge-pathnames *LispWeb-init-file* *LispWeb-conf-dir*)))
    (http-log-message "Loading initialization file ~A" path)
    (load path :verbose nil :if-does-not-exist nil)))

(defun load-personal-init-file ()
  (let ((path (merge-pathnames (format nil ".~a" *LispWeb-init-file*) 
			       *home-user-pathname*)))
    (When (probe-file path)
      (http-log-message "Loading initialization file ~A" path)
      (load path :verbose nil :if-does-not-exist nil))))
      
;;; Main server function

(defun start-tcp-services (&optional options backgroundp)
  "Start the TCP services provided by the LispWeb server (normally, 
an HTTP server and a Telnet server)."
  (declare (ignore backgroundp))
  (init-server options)			;load init file, handle command-line arguments & open streams
  #+harlequin-unix-lisp
  (when *eval-port*
    (start-telnet-server))
  (when *http-port*
    (lisp-http-server)))

(defun cl-user::start-tcp-services (&optional options backgroundp)
  (start-tcp-services options backgroundp))

(defun start-tcp-services-with-mp (&optional options backgroundp)
  "Start the TCP services provided by the LispWeb server (normally, 
an HTTP server and a Telnet server)."
  (declare (ignore backgroundp))
  (init-server options)			;load init file, handle command-line arguments & open streams
  #+harlequin-unix-lisp
  (when *eval-port*
    (tcp::start-telnet-server-with-mp))
  (when *http-port*
    (tcp::lisp-http-server-with-mp)))

;;; Misc functions

(defun lispweb-path (filename)
  "Given an absolute or relative pathname string, generate a pathname relative to *LispWeb-root-dir*"
  (when (and (not (string= filename ""))
	     (or (eq (aref filename 0) #\/)
                 ;;for pcs john domingue march 98
                 (eq (aref filename 0) #\\)))
    (setq filename (subseq filename 1)))
  (format nil "~a" (merge-pathnames filename *LispWeb-root-dir*)))
  
(defun encrypt-password (password)
  "Encrypt the initial password"
  (crypt password (format nil "~a~a" 
			  ;;(coerce (+ 65 (random 26)) 'character) 
			  ;;(coerce (+ 65 (random 26)) 'character))))
                          ;;;for pc
                          (code-char (+ 65 (random 26)))
                          (code-char (+ 65 (random 26))))))

;;;for pc 

(defun crypt (password x)
  password)


(defun check-password (password)
  "Check the supplied password against the true one."
  (string= *operator-password*
	   (crypt password
		  (subseq *operator-password* 0 2))))

(defun http-time (&optional (gmt-p t))
  "Return the current date in RFC-1123 format."
  (let ((days '("Mon" "Tue" "Wed" "Thu" "Fri" "Sat" "Sun"))
	(months '(nil "Jan" "Feb" "Mar" "Apr" "May" "Jun"
		      "Jul" "Aug" "Sep" "Oct" "Nov" "Dec")))
    (multiple-value-bind
	  (second minute hour date month year day-of-week dst-p tz)
	(get-decoded-time)
      (cond (gmt-p
	     (setq hour (+ hour tz (if dst-p 1 0)))
	     (format nil "~a, ~2,'0d ~a ~d ~2,'0d:~2,'0d:~2,'0d GMT"
		     (nth day-of-week days)
		     date (nth month months) year
		     hour minute second))
	    (t (if dst-p (incf hour))
	       (format nil "~a, ~2,'0d ~a ~d ~2,'0d:~2,'0d:~2,'0d"
		       (nth day-of-week days)
		       date (nth month months) year
		       hour minute second))))))

(defun short-date ()
  (let ((date (multiple-value-list (get-decoded-time))))
    (format nil "~d/~2,'0d/~2,'0d" (nth 3 date) (nth 4 date) (nth 5 date))))

(defun short-time ()
  (let ((date (multiple-value-list (get-decoded-time))))
    (format nil "~d:~2,'0d:~2,'0d" (nth 2 date) (nth 1 date) (car date))))

(defun splice-string (string separators)
  (let ((start 0)
	(end 0)
	(result nil))
    (dolist (s separators)
      (setq end (position s string :start start))
      (unless end			;is this enough?
	(return))
      (push (subseq string start end) result)
      (setq start (1+ end)))
    (reverse result)))
  
(defun safe-readnum (string &optional (default 0))
  (or (parse-integer string :junk-allowed t)
      default))

(defun http-to-universal-time (time)
  "Time is a string in RFC-1123 format."
  (let ((decoded (splice-string time '(#\, #\Space #\Space #\Space #\Space #\: #\: #\Space)))
	(months '(nil "Jan" "Feb" "Mar" "Apr" "May" "Jun" "Jul" "Aug" "Sep" "Oct" "Nov" "Dec")))
    (let ((date (read-from-string (third decoded)))
	  (month (position (fourth decoded) months :test #'string=))
	  (year (read-from-string (fifth decoded)))
	  (hour (read-from-string (sixth decoded)))
	  (minute (read-from-string (seventh decoded)))
  	  (second (read-from-string (eighth decoded))))
      (encode-universal-time second minute hour date month year))))

;;; Requests

(defvar *xml-type-string* "text/xml")

(defun xml-type-p (x)
  (when (listp x)
    (setf x (car x)))
  (search *xml-type-string* x))

(defun not-xml-type (headers)
  (let ((value t))
    (mapc #'(lambda (header)
              (when (and (eq (header-field-name header) :content-type)
                         (xml-type-p (header-field-value header)))
                (setf value nil)))
          headers)
    value))

(defun http-read-line (stream &optional eof)
  ;;(format t "~%before read-line in http-read-line")
  (let ((line (remove #\Return (remove #\null (read-line stream nil eof)))))
    ;;(format t "~%after read-line in http-read-line")
    (lw-log :http-debug "HTTP < ~s~%" line)
    line))

(defun http-force-output (stream)
  (when (output-stream-p stream)
    (force-output stream)))

(defun no-break (&rest ignore)
  (declare (ignore ignore)))

(defun get-content-length (info)
  (let ((headers (http-info-headers info)))
    (dolist (header headers)
      (when (eq (header-field-name header) :content-length)
        (return (read-from-string (car (header-field-value header))))))))

(defun process-request (info)
  ;;(setf xii info)
  (let ((*http-info* info)
	(stream (http-info-stream info))
        (break-fun (symbol-function 'break)))
    (unwind-protect
	(let (method)
          ;;(format t "~%before proces request line")
	  (process-request-line)
	  (setq method (http-info-method *http-info*))
          ;;(setf mm method r *http-info*)
	  (unless (eq (http-info-protocol *http-info*) :http-0.9) ;HTTP/0.9 has no headers
	    (setf (http-info-headers *http-info*)
	          (process-headers stream)))
	  
	  (setf (http-info-browser *http-info*)
	        (call-handler *LispWeb-browser-table* (header-value :user-agent)))
	  
          ;;; (A bit less) horrible kludge to handle jave class requests ! 
	  (let* ((req-obj (http-info-request-object *http-info*))
		 (req-len (length req-obj)))
	    (when (string= (subseq req-obj (max 0 (- req-len 6))) ".class")
	      (setf (http-info-request-object *http-info*)
					;(concatenate 'string "/class" req-obj))))
		    (applet-path req-obj))))
	  
	  (when (eq method :post)
            (if (not-xml-type (http-info-headers *http-info*))
	        (setf (http-info-fields *http-info*)
	              (process-form-fields stream))
              (setf (http-info-fields *http-info*)
	              (process-xml-input stream))))
	  
	  ;;; Info fields are now set. Bind the output stream and dispatch to the appropriate method function.
	  (let ((*http-stream* (http-info-stream *http-info*))
		(request (http-info-request-object *http-info*))
		(start-time (get-internal-run-time)))
	    (unless (eq (http-info-protocol *http-info*) :invalid)
	      (cond (*LispWeb-debug*
		     (http-reply method request))
		    (t
                     #+lispworks (sys::without-warning-on-redefinition
				   (setf (symbol-function 'break)
				         (symbol-function 'no-break)))
		     #+allegro (excl:without-package-locks
				(setf (symbol-function 'break)
				      (symbol-function 'no-break)))
		     
		     (handler-case
			 (http-reply method request)
		       ;;; Default condition handlers
		       (stream-error (c) ;connection interrupted
			             (declare (ignore c))
			             (http-debug-message "Connection interrupted")
			             nil)
		       (file-error (c)	;connection interrupted
			           (declare (ignore c))
			           (http-debug-message "File error")
			           nil)
		       (no-authorization (c) ;no authorization
			                 (declare (ignore c))
			                 (no-auth-page request)
			                 (http-debug-message "No authorization"))
		       (bad-access-key (c)
			               (declare (ignore c))
			               (bad-key-page request)
			               (http-debug-message "Bad page key"))
		       (access-key-expired (c)
			                   (declare (ignore c))
			                   (key-expired-page request)
			                   (http-debug-message "Page key expired"))
		       (head-request (c) ;head request - do nothing
			             (declare (ignore c)))
		       (error (c)	;generic error (should not occurr!)
					;(setq *debug-io* (http-info-stream *http-info*))
					;(error c)
			      (error-page c)
			      (http-debug-message "Error ~s" c)))))
              (handler-case
	          (http-force-output *http-stream*)
                (stream-error (c) ;connection interrupted
			      (declare (ignore c))
			      (http-debug-message "Connection interrupted")
			      nil)
                (file-error (c)	;connection interrupted
			    (declare (ignore c))
			    (http-debug-message "File error")
			    nil)
                (error (c)
                       (http-debug-message "Connection interrupted - reason unknown" c)
                       nil)))
	    (setf (http-info-time *http-info*)
	          (- (get-internal-run-time) start-time))
	    ;;taken out by john domingue 31/5/02
            ;;(http-log *http-info*)
            ))
      (handler-case
          (http-force-output stream)
        (stream-error (c) ;connection interrupted
                      (declare (ignore c))
                      (http-debug-message "Connection interrupted")
                      nil)
        (file-error (c)	;connection interrupted
                    (declare (ignore c))
                    (http-debug-message "File error")
                    nil)
        (error (c)
               (http-debug-message "Connection interrupted - reason unknown" c)
               nil))
      #+harlequin (sys::without-warning-on-redefinition
		    (setf (symbol-function 'break) break-fun))
      #+allegro (excl:without-package-locks
		 (setf (symbol-function 'break) break-fun))
      )
    )
  )
#|
      (progn
	(force-output stream)		;ensure output is forced...
	(sleep 1)			;wait a little bit...
	;(tcp::close-tcp-stream stream)  ;close stream at the end
	(remove-lispweb-process (http-info-request-counter *http-info*)))))) ; and remove process
	|#

;;; Headers

(defun process-request-line ()
  (let* ((stream (http-info-stream *http-info*))
	 (request-line (http-read-line stream "NIL"))
	 (method (unless (zerop (length request-line))
                   (let ((*package* (find-package :keyword)))                       
                     (read-from-string request-line)))))
    (http-log-message "~%process-request-line: ~a" request-line)
    (unless (or (zerop (length request-line)) (string= request-line "NIL"))
      (setf (http-info-request-line *http-info*) request-line)
      (setf (http-info-method *http-info*) method)
      
;;; Requests can be in one of the following formats:
;;; GET object	                 (http/0.9)
;;; method object HTTP/1.0	 (http/1.0)	
      
      (setq request-line (string-trim '(#\Space) request-line)) ;ensure no spaces before and after
      
      (let ((first (position #\Space request-line))
	    (last (position #\Space request-line :from-end t)))
	(cond ((null first)		;no spaces, invalid request
	       (setf (http-info-protocol *http-info*) :invalid))
	      ((= first last)		;one space
	       (setf (http-info-protocol *http-info*) :http-0.9)
	       (setf (http-info-request-object *http-info*) 
		 (subseq request-line (1+ first))))
	      ((string= (subseq request-line (1+ last)) "HTTP/1.0") ;two spaces
	       (setf (http-info-protocol *http-info*) :http-1.0)
	       (setf (http-info-request-object *http-info*) 
		 (subseq request-line (1+ first) last)))
	      (t (setf (http-info-protocol *http-info*) :other)
		 (setf (http-info-request-object *http-info*) 
		   (subseq request-line (1+ first) last))))))))

(defun process-response-line ()
  (let* ((stream (http-info-stream *http-info*))
	 (response-line (http-read-line stream "NIL")))
    (unless (string= response-line "NIL")
      (setf (http-info-request-line *http-info*) response-line)
      
;;; Responses can be in one of the following formats:
;;; Protocol Code 
;;; Protocol Code Phrase
      
      (setq response-line (string-trim '(#\Space) response-line)) ;ensure no spaces before and after
      (let ((first (position #\Space response-line))
	    (last (position #\Space response-line :from-end t)))
	(cond ((null first)		;no spaces, invalid response
	       (setf (http-info-protocol *http-info*) :invalid))
	      ((= first last)
	       (setf (http-info-protocol *http-info*)
		 (intern (subseq response-line 0 first) :keyword))
	       (setf (http-info-code *http-info*)
		 (safe-readnum (subseq response-line last) 500))) ; default code is server error
	      (t (setf (http-info-protocol *http-info*)
		   (intern (subseq response-line 0 first) :keyword))
		 (setf (http-info-code *http-info*)
		   (safe-readnum (subseq response-line first last) 500))))))))

(defun process-headers (stream)
  (let ((headers nil)
	(line nil)
	(colon nil)
	(semicolon nil)
	(key nil)
	(value nil)
	(options nil)
	(header nil))
    (loop
     ;;(format t "~%process-headers: before http-read-line")
      (setq line (http-read-line stream ""))
      ;;(format t "~%process-headers: header line = ~s" line)
      (when (string= "" line)
	(return))
      (when (setq colon (position #\: line)) ;when syntax is correct...
	(setq key (intern (string-upcase (subseq line 0 colon)) :keyword))
	(setq header (find key headers :key #'header-field-name)) ;find or create the header
	(unless header
	  (setq header (make-header-field :name key))
	  (push header headers))
	(if (setq semicolon (position #\; line)) ; if semicolon is present
	    (progn 
	      (setq value (subseq line (+ colon 2) semicolon)
		    options (parse-header-options (subseq line (+ semicolon 2)))) ;try to parse options
	      (when (member :none options) ;if options are invalid
		(setq value (subseq line (+ colon 2))
		      options nil)))	;ignore semicolon
	  (setq value (subseq line (+ colon 2))
		options nil))
	
	(push value (header-field-value header))
	(when options 
	  (push options (header-field-options header)))))
    headers))

(defun make-option-pair (string)
  (let ((equal (position #\= string)))
    (if equal
	(list (intern (string-upcase (subseq string 0 equal)) :keyword)
	      (subseq string (1+ equal)))
      (list :none))))

(defun parse-header-options (line &aux semicolon)
  (if (setq semicolon (position #\; line))
      (append (make-option-pair (subseq line 0 semicolon))
	      (parse-header-options (subseq line (+ semicolon 2))))
    (make-option-pair line)))

;;; Header fields access

(defun find-header (name)
  (find name (http-info-headers *http-info*) 
	:key #'header-field-name))
  
(defun header-values (name)
  "Retrieve all the values of the specified header."
  (let ((hdr (find-header name)))
    (and hdr (header-field-value hdr))))

(defun header-value (name)
  "Retrieve the first value of the specified header."
  (car (header-values name)))

;;(defun add-header-value (name value)
;;  (push value (find-header name)))

;;;chnaged by john for pc lispworks april '98
(defun add-header-value (name value)
  (let ((l (find-header name)))
    (push value l)))

(defun set-header-values (name values)
  (setf (header-field-value (find-header name)) values))

(defun print-header-field (hdr)
  (format nil "~:(~a~): ~{~a~^,~}~:[~; ~:*~a~]"
	  (header-field-name hdr)
	  (header-field-value hdr)
	  (header-field-options hdr)))

;;; Form fields

;;;original process-form-fields
#|
(defun process-form-fields (stream)
  (let ((line (http-read-line stream nil))
	(fields nil))
    (declare (special fields))
    (when (and line (not (zerop (length line))))
      (process-form-fields-aux line))
    fields))
|#

(defun http-read-chars (stream length eof)
  (let ((line nil))
    (dotimes (i length)
      (push (read-char stream nil eof) line))
    (coerce (reverse line) 'string)))

(defun process-form-fields (stream)
  (let* ((length (get-content-length *http-info*))
        (line (http-read-chars stream length nil))
	(fields nil))
    (declare (special fields))
    (when (and line (not (zerop (length line))))
      (process-form-fields-aux line))
    fields))

(defun process-xml-input (stream)
  (let* ((length (get-content-length *http-info*))
        (line  (http-read-chars stream length nil))
        (lines nil))
    (when (and line (not (zerop (length line))))
      (push line lines))
    (apply #'concatenate 'string (reverse lines))))

(defun process-form-fields-aux (request)
  (declare (special fields))
  (let ((mark1 (position #\= request))
	(mark2 (position #\& request)))
    (let ((field-name (intern (string-upcase (translate-request (subseq request 0 mark1))) :keyword))
	  (field-value (translate-request (subseq request (1+ mark1) mark2))))
      (if (getf fields field-name)
	  (setf (cdr (last (getf fields field-name)))
	    (cons field-value nil))
	(setf (getf fields field-name) (list field-value)))
      (if mark2 (process-form-fields-aux (subseq request (1+ mark2)))))))

;;; Form fields access

(defun form-value (name)
  "Retrieve the value of the form field called name."
  (car (getf (http-info-fields *http-info*) name)))

(defun form-values (name)
  "Retrieve the values of the form fields called name."
  (getf (http-info-fields *http-info*) name))

;;; Form field setting

(defun add-form-value (name value)
  (push value (getf (http-info-fields *http-info*) name)))

(defun set-form-values (name values)
  (setf (getf (http-info-fields *http-info*) name) values))

;;; Utilities

(defun explode-request (request &optional (separator #\/))
  "Turn a slash-separated string into a list of strings."
  (unless (string= request "")
    (if (eq (aref request 0) separator)
	(explode-request (subseq request 1) separator)
      (let ((pos (position separator request :start 1)))
	(if pos
 	    (cons (subseq request 0 pos)
		  (explode-request (subseq request (1+ pos)) separator))
	  (list request))))))

(defun translate-request (request)
  (let ((mode nil)
	(hex 0))
    (with-output-to-string (result)
      (every
       #'(lambda (char)
	   (cond ((eq mode :first)
		  (setq hex (let ((*read-base* 16)) 
			      (read-from-string (format nil "~a" char))))
		  (setq mode :second))
		 ((eq mode :second)
		  (setq hex (+ (* 16 hex) 
			       (let ((*read-base* 16)) 
				 (read-from-string (format nil "~a" char)))))
                  #+lispworks3.2.2
		  (format result "~a" (coerce hex 'character))
                  #+lispworks4
		  (format result "~a" (code-char hex))
		  (setq mode nil)
		  t)
		 ((eq char #\+)
		  (format result " ")
		  t)
		 ((eq char #\%)
		  (setq mode :first)
		  t)
		 (t (format result "~a" char)
		    t)))
       request)
      result)))

;;; Condition for unauthorized access

(define-condition no-authorization
    (condition)
  nil)

(define-condition bad-access-key
    (condition)
  nil)

(define-condition access-key-expired
    (condition)
  nil)

(defvar *no-auth-condition*
    (make-condition 'no-authorization))

(defvar *bad-key-condition*
    (make-condition 'bad-access-key))

(defvar *key-expired-condition*
    (make-condition 'access-key-expired))

;;; Authorization management

(defun make-auth-list (host)
  "Take a string representing a host address and turn it into a list that can be used 
in the authorization list of a page."
  (let ((*readtable* (copy-readtable)))
    (set-syntax-from-char #\. #\Space)
    (read-from-string (concatenate 'string "(" host ")"))))

(defun match-auth (a1 a2)
  "Match two host addresses."
  (cond ((and (null a1)
	      (null a2))
	 t)
	((or (eq (car a1) '*)
	     (eq (car a2) '*))
	 t)
	((= (car a1)
	    (car a2))
	 (match-auth (cdr a1) (cdr a2)))
	(t nil)))

(defun test-auth (host auth-list)
  (some #'(lambda (auth)
	    (match-auth host auth))
	auth-list))

(defun check-auth (auth-list)
  (let ((this (make-auth-list (http-info-addr *http-info*))))
    (unless (test-auth this auth-list)
      (error *no-auth-condition*))))

;;; Access key management

(defun test-access-key (string)
  (let ((result (check-key string)))
    (cond ((stringp result)
	   result)
	  ((eq result :expired)
	   (error *key-expired-condition*))
	  (t (error *bad-key-condition*)))))

;;; HTTP Client

(defun http-client (&key host (port 80) (request "/") (stream t))
  "Send a request to the specified HTTP server. Keywords:
host - the HTTP server
port - the HTTP port (default is 80)
request - the requested URL (default is /)."

  (let ((stm (telnet host port)))
    (format stm "GET ~a HTTP/1.0~%~%" request)
    (http-force-output stm)
    (handler-case
	(loop
	  (format stream "~a~%" (read-line stm)))
      (end-of-file (c)
	(declare (ignore c))
	(print "Connection closed.")))))

