;;; -*- Mode: LISP; Syntax: Common-lisp; Base: 10; Package: OCML;   -*-

;;;This file contains the definitions linking the generic problem
;;;solving model to specific design knowledge.

(in-package "OCML")

(defvar *design-constraints* nil)
(defvar *design-parameters* nil)
(defvar *design-requirements* nil)
(defvar *design-operators* nil)
(defvar *cost-function* nil)
(defvar *cost-order-rel* nil)
(defvar *cost-difference-fun* nil)
(defvar *cost-sum-fun* nil)
(defvar *assignment-rel* nil)
(defvar *trace-design-process?* t)
;;;;;(defvar *compute-cost?* t)
(defvar *current-state* nil)
(defvar *state-counter* 0)
(defvar *current-method*)
(defvar *current-task*)


(defun cache-design-specification (appl method)
  (let ((task (the-slot-value appl 'tackles-task)))
  (setf *current-task* (find-instance 'task)
          *current-method* (find-instance 
                            (the-slot-value *current-application* 'uses-method) 
                            'problem-solving-method)
          *design-operators* 
          (mapcar #'(lambda (x)
                      (find-instance x 'design-operator))
                  (the-slot-value (name method) 'has-design-operators))
          *design-parameters* 
          (mapcar #'(lambda (x)
                      (find-instance x 'parameter))
                  (the-slot-value task 'has-parameters))
          *design-constraints* 
          (mapcar #'(lambda (x)
                      (find-instance x 'constraint))
                  (the-slot-value task 'has-constraints))
          *design-requirements* 
          (mapcar #'(lambda (x)
                      (find-instance x 'requirement))
                  (the-slot-value task 'has-requirements))
          *assignment-rel* (the-slot-value 
                            (the-slot-value task 'has-design-model)
                            'membership-test)
          *cost-function* (the-slot-value task
                                          'has-cost-function)
          *cost-difference-fun* (second (the-slot-value task
                                         'has-cost-algebra))
          *cost-sum-fun* (first (the-slot-value task
                                         'has-cost-algebra))
          *cost-order-rel* (third (the-slot-value task
                                         'has-cost-algebra)))))

(defun construct-dependency-network ()
  (loop for param in *design-parameters*
        for depends-on = (get-slot-values param 'depends-on)
        do
        (loop for el in depends-on
              do
              (add-slot-value el 'affects (name param)))))
      

(defun clear-cached-links ()
  (loop for op in *design-operators*
        do
        (setf ( effective-parameter-list op) nil))
  (loop for p in *design-parameters*
        do
        (setf (relevant-operators p) nil)
        (setf (relevant-constraints p) nil))
   (loop for c in *design-constraints*
        do
        (setf (effective-parameter-list c) nil)))
  


;;;CACHE-PARAM-TO-OPS-LINKS
(defun cache-param-to-ops-links ()
  (loop for op in *design-operators*
        for params = (compute-effective-applicability-set op)
        do
        (loop for param in params
              do
              (link-param-to-operator (find-instance param 'parameter)
                                      op))))

(defun cache-param-to-constraints-links ()
  (loop for c in *design-constraints*
        for params = (compute-effective-applicability-set c)
        do
        (loop for param in params
              do
              (link-param-to-constraint (find-instance param 'parameter)
                                      c))))

(defmethod LINK-PARAM-TO-OPERATOR ((param parameter) (op design-operator))
    (with-slots (effective-parameter-list) op
      (pushnew param effective-parameter-list)
      (pushnew op  (relevant-operators param))))

(defmethod LINK-PARAM-TO-CONSTRAINT ((param parameter) (c constraint))
    (with-slots (effective-parameter-list) c
      (pushnew param effective-parameter-list)
      (pushnew c  (relevant-constraints param))))



;;;*DESIGN-SPACE*
(defvar *design-space* nil)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;APPLY-METHOD - An application is executed by applying the chosen method 
;;;to the instance representing the application.  This definition ensures that for
;;;design methods this call results in a call to top-level-design.
(defmethod APPLY-METHOD ((method design-method) appl)
  (top-level-design appl method ))

(defmethod TOP-LEVEL-DESIGN (appl (method design-method))
  (clear-cached-links)
  (cache-design-specification appl method)
  (cache-param-to-ops-links)
  (cache-param-to-constraints-links)
  (construct-dependency-network)
  (initialise-design-space appl method)
  (top-level-design-loop appl method))

(defmethod TOP-LEVEL-DESIGN-LOOP (appl (method design-method))
  (catch 'exit
    (loop
      (let ((state (select-design-state appl method *design-space*)))
        (cond (state
               (make-this-state-current appl state method)
               (design-from-state appl method state))
              (t (design-fails)))))))


;;;METHOD INITIALISE-DESIGN-SPACE
;;;Generates the starting point in the search space
(defmethod initialise-design-space (appl (method design-method))
  (setf *design-space* nil)
  (setf *current-state* nil)
  (setf *state-counter* 0)
  (add-design-state appl method nil nil nil))

(defun INITIAL-STATE ()
  (car *design-space*))


(defmethod ADD-DESIGN-STATE (appl (method design-method) predecessor 
                                  extensions removals)
  (let ((state (new-design-state appl method predecessor extensions removals)))
    
    (internal-add-state method state)
    (evaluate-state appl method state)))

(defmethod REMOVE-DESIGN-STATE ((state design-state))
  (setf *design-space*
        (remove state *design-space*)))

(defmethod NEW-DESIGN-STATE (appl (method design-method)
                                  &optional  predecessor extensions removals)
 
  appl
  (cond (predecessor
         (make-instance (design-state-type method)
           :design-model (set-difference 
                          (append extensions (design-model predecessor))
                          removals
                          :test #'equal)
           :design-extensions extensions
           :design-removals removals
           :depth (1+ (state-depth predecessor))
           :predecessor predecessor
          ))
        (t
         (make-instance (design-state-type method)
           :design-model nil))))


(defmethod internal-add-state ((method design-method) (state design-state))
  (incf *state-counter*)
  (setf *design-space* (cons state *design-space*)))



(defmethod MAKE-THIS-STATE-CURRENT (appl (state design-state)(method design-method))
  appl
  (unless (eq state *current-state*)
    (setf *current-state* state)
    (with-slots (design-model) state
      (loop for pair in (setofall '(?x . ?C)`(,*assignment-rel* ?x ?C ))
            unless (member pair design-model :test #'equal)
            do
            (unassert1 `(,*assignment-rel* ,(car pair) ,(cdr pair))))
      (loop for assignment in design-model
            do
            (tell1 `(,*assignment-rel* ,(car assignment) ,(cdr assignment)))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defmethod SELECT-DESIGN-STATE (appl (method design-method) space)
  appl method

  (when *trace-design-process?*
    (when space
      (lines 2)
      (print-with-spaces 2 ;;; 2 ;;;  (state-depth (car space)) 
                         "Current state is now ~s"
                         (car space))))

  (car space))

(defmethod SELECT-DESIGN-FOCUS (appl (method design-method) (state design-state) 
                                     context)
     appl method context
     (car (state-foci state)))


(defmethod SELECT-DESIGN-FOCUS (appl (method design-method) (state design-state) 
                                     (context (eql :extend)))
     appl method
     (when (state-foci state)
       (multiple-value-bind (focus winners)
                            (select-most-restricted-parameter 
                             (mapcar #'name 
                                     (state-foci state)))
         (setf focus 
               (select-most-influential-parameter 
                winners
                #'(lambda (x)
                    (setofall '?x `(depends-on ,x ?x)))
                #'(lambda (x)
                    (setofall '?x `(affects ,x ?x)))))
         (when *trace-design-process?*
           (if focus
             (print-with-spaces 2 ;;; 2 ;;; (state-depth state)
                                "Focus on ~a" focus)
             (print-with-spaces 2 ;;; (state-depth state)
                                "Cannot select any focus")))
         (when focus
           (find-instance focus)))))


;(defun select-most-influential-parameter (parameters)
;  (Let* ((alist (sort (mapcar #'(lambda (x)
;                                 (cons x (length (setofall '?x `(depends-on ,x ?x)))))
;                            parameters)
;                     #'(lambda (x y)
;                         (> (cdr x)(cdr y)))))
;        (winners (filter-ordered-list alist #'(lambda (pair)
;                                                (= (cdr pair)
;                                                   (cdar alist))))))
;    (setf alist (sort winners
;                      #'(lambda (x y)
;                          (> (length (setofall '?x `(affects ,(car x) ?x)))
;                             (length (setofall '?x `(affects ,(car y) ?x)))))))
;   (caar alist)))


;;;This definition avoids determinism in the choice of parameters
(defun select-most-influential-parameter (parameters influences-in-fun influences-out-fun)                                          
  (Let* ((alist (sort (mapcar #'(lambda (x)
                                  (cons x (length (funcall influences-in-fun x))))
                              parameters)
                      #'(lambda (x y)
                          (> (cdr x)(cdr y)))))
         (winners (filter-ordered-list alist #'(lambda (pair)
                                                 (= (cdr pair)
                                                    (cdar alist))))))
    (setf alist (sort (mapcar #'(lambda (x)
                                  (cons (car x) 
                                        (length 
                                         (funcall influences-out-fun (car x)))))
                              winners)
                      #'(lambda (x y)
                          (> (cdr x)(cdr y))))
          winners (filter-ordered-list alist #'(lambda (pair)
                                                 (= (cdr pair)
                                                    (cdar alist)))))
    (values 
     (car (elt winners (random (length winners))))
     (mapcar #'car winners))))

(defun select-most-restricted-parameter (parameters)
  (when parameters
    (loop for p in (cdr parameters)
          with winners = (list (car parameters))
          with value = (length (the-slot-value (car parameters) 'has-value-range)) 
          for l = (length (the-slot-value p 'has-value-range))
          do
          (cond ((= l value)
                 (push p winners))
                ((< l value)
                 (setf winners (list p)
                       value l)))
                      
          finally
          (return 
           (values (elt winners (random (length winners)))
                   winners)))))
        
                    

(defmethod EVALUATE-STATE (appl (method design-method)(state design-state))
  (if (eq *current-state* state)
    (evaluate-state-internal appl method state)
      (Let ((s *current-state*))
        (make-this-state-current appl state method)
        (evaluate-state-internal appl method state)
        (when s
          (make-this-state-current appl s method)))))

(defmethod EVALUATE-STATE-INTERNAL (appl (method design-method)(state design-state))
   (setf (state-feasible? state) (evaluate-feasibility appl method state)
        ;;;;(state-foci state) (collect-state-parameters appl state)
      ;;;;  (state-operators state)(collect-applicable-operators appl method state) 
        (violated-constraints state) (evaluate-constraints appl method
                                                           state)
        (state-complete? state) (design-complete? appl method state)
        ;;;(state-cost state) (evaluate-state-cost appl method state)
        ))


(defmethod EVALUATE-FEASIBILITY (appl (method design-method)(state design-state))
  (declare (ignore appl method state))
  "This is just to provide a default"
  t)

(defmethod EVALUATE-CONSTRAINTS (appl (method design-method)(state design-state))
  appl
  (when (and *trace-design-process?* *design-constraints*)
     (print-with-spaces 2 ;;; (state-depth state)
                        "Evaluating constraints....."))
  (with-slots (design-extensions bad-assignments) state
    (loop with constraints-violated
          for pair in design-extensions
          for param = (car pair)
          for value = (cdr pair)
          for relevant-constraints = (relevant-constraints (find-instance param))
          do
          (loop for c in relevant-constraints
                for flag = (constraint-precondition-holds? c param value)
                when flag
                do
                (setf flag (assignment-violates-constraint? c param value))
                (when flag
                  (pushnew c constraints-violated)
                  (push (cons  c (cons param value)) bad-assignments)))
          finally
          (if ;;;;;(and *trace-design-process?*
                   constraints-violated ;;;;)
            (print-with-spaces 2 ;;; (state-depth state)
                               "The following constraints have been violated: ~{~A ~}"
                               constraints-violated)
            (when *trace-design-process?*
              (print-with-spaces 2 ;;; (state-depth state)
                                 "...Done")))
      (return constraints-violated))))



(defmethod FILTER-APPLICABLE-OPERATORS ((state design-state) ops)
  (loop with applicable-operators
        for op in ops
        for result = (precondition-holds? op (design-model state))
        when result
        do
        (push op applicable-operators)
        (setf (applicability-env op) (when (listp result) result))
        finally
        (return  applicable-operators)))

(defmethod COLLECT-OPERATORS-APPLICABLE-TO-FOCUS (appl (method design-method)
                                                       (state design-state)
                                                       focus
                                                       context)
  appl context
  (filter-applicable-operators 
   state
   (relevant-operators focus)))


(defmethod COLLECT-APPLICABLE-OPERATORS (appl (method design-method)
                                              (state design-state))
  appl
  (filter-applicable-operators 
   state
   (remove-duplicates (apply #'append 
                             (mapcar #'relevant-operators 
                                     (state-foci state))))))

(defmethod COLLECT-STATE-PARAMETERS (appl (state design-state))
  appl
  (all-unbound-computable-parameters *design-parameters* (design-model state)))


;(defmethod DESIGN-COMPLETE? (appl (method design-method)
;                                  (state design-state))
;  appl
;  (holds? `(design-model-complete ,(design-model state)
;                                  ,(mapcar #'name
;                                        *design-parameters*))))

;;;DESIGN-COMPLETE? --This is more efficient than the one above
(defmethod DESIGN-COMPLETE? (appl (method design-method)
                                  (state design-state))
  appl
  (= (length (design-model state) ) (length *design-parameters*)))

    
        
(defun DESIGN-SUCCEEDS (state)
  state
  (print-with-spaces 2 ;;; (state-depth state) 
                     "Design is now complete.  Cost of final state is ~s"
                     (state-cost state))
  
  (throw 'exit t))

(defun DESIGN-FAILS (&optional state)
  state
  
  (print-with-spaces 2 ;;; (if state (state-depth state) 0)
                       "Cannot select a state...design fails")

  (throw 'exit nil))

(defmethod state-fails ((state design-state))
  (when *trace-design-process?*
    (print-with-spaces 2 ;;; (state-depth state) 
                       "State ~s fails" state))
  (setf (state-failed?  state) t)
  (remove-design-state state))

(defmethod state-fully-expanded ((state design-state))
  (when *trace-design-process?*
    (print-with-spaces 2 ;;; (state-depth state) 
                       "State ~s fully expanded" state))
  ;;;;(setf (state-failed?  state) t)
  (remove-design-state state))
  

(defmethod EXTEND-DESIGN (appl (method design-method) (state design-state))  
  (cond ((new-state? state)
         (setf (state-foci state) (collect-state-parameters appl state)
               (new-state? state) nil)
         (design-from-context appl method state :extend))
        (t
         (retry-design-operator appl method 
                                (current-operator state) state 
                                (current-focus state) :extend))))

(defmethod DESIGN-FROM-CONTEXT (appl (method design-method) (state design-state)
                                     context)
  
  (let ((focus (select-design-focus appl method state context)))
    (cond (focus
           (setf (current-focus state) focus
                 (state-foci state) (remove focus 
                                            (state-foci state))
                 (focus-operators state) 
                 (sort-operators appl method state
                                 (collect-operators-applicable-to-focus 
                                  appl method state focus context)))
           (cond ((focus-operators state)
                  (design-from-focus appl method state focus context))
                 (t
                  (backtrack-from-focus appl method state focus context))))
          (t 
           (state-fully-expanded state)))))
  


(defmethod DESIGN-FROM-FOCUS (appl (method design-method)  
                                   (state design-state)
                                   focus
                                   context)
  (let ((design-op (select-design-operator appl method state focus 
                                           (focus-operators state))))
    (cond (design-op
           (setf (current-operator state) design-op
                 (nogood-values state) nil
                 (focus-operators state) 
                 (remove 
                  design-op 
                  (focus-operators state)))
           (try-design-operator appl method design-op state focus context))
          (t
           (backtrack-from-focus appl method state focus context)))))


(defmethod BACKTRACK-FROM-FOCUS (appl 
                                 (method design-method) 
                                 (state design-state) 
                                 focus
                                 context)
  (default-backtrack-from-focus-actions appl method state focus context))


(defun DEFAULT-BACKTRACK-FROM-FOCUS-ACTIONS (appl method state focus context)
  (when *trace-design-process?*
    (print-with-spaces 2 ;;; (state-depth state)
                       "No more operators applicable to ~a...trying another focus" 
            (name focus)))
  (design-from-context appl method state context))


(defmethod FILTER-OPERATORS-APPLICABLE-TO-FOCUS ((method design-method) 
                                                (state design-state)  
                                                (focus parameter))
  (intersection (state-operators state)
                (relevant-operators focus)))


(defmethod SORT-OPERATORS (appl (method design-method) 
                                (state design-state)
                                ops)
  appl
  (mapcar #'find-instance
          (ocml-eval-fun `(sort ',(mapcar #'name ops)
                                design-operator-order))))

(defmethod SELECT-DESIGN-OPERATOR (appl (method design-method) 
                                   (state design-state)
                                   focus
                                   ops)
  (declare (ignore   focus method appl))
  (let ((op (car ops)))
    (when *trace-design-process?*
      (if op
        (print-with-spaces 2 ;;; (state-depth state)
                           "Current operator is ~a" (name op))
        (print-with-spaces 2 ;;; (state-depth state)
                           "Cannot select an operator")))
    op))

(defmethod RETRY-DESIGN-OPERATOR (appl (method design-method)
                                       (design-op basic-design-extension-operator)
                                       (state design-state)  
                                       focus
                                       context)
 
  (when *trace-design-process?*
    (print-with-spaces 2 ;;; (state-depth state)
                       "Retrying ~a on ~a" 
                       (name design-op)
                       (name focus)))
  (backtrack-from-operator appl method state design-op focus context))

(defmethod RETRY-DESIGN-OPERATOR (appl (method design-method)
                                       (design-op multiple-design-extension-operator)
                                       (state design-state)  
                                       focus
                                       context)
 
  (when *trace-design-process?*
    (print-with-spaces 2 ;;; (state-depth state)
                       "Retrying ~a on ~a" 
                       (name design-op)
                       (name focus)))
  (try-design-operator appl method design-op state focus context))
 
(defmethod TRY-DESIGN-OPERATOR (appl (method design-method)
                                     (design-op basic-design-extension-operator)
                                     (state design-state)  
                                     focus
                                     context)
  (let ((value (apply-design-operator 
                design-op state  focus context)))
  (cond ((eq value :nothing)
         (backtrack-from-operator appl method state design-op focus context))
        (t
         (when *trace-design-process?*
           (print-with-spaces 2 ;;; (state-depth state)
                              "New assignment is <~a ~A>" (name focus) value))
         (add-design-state appl  method state  (List (cons (name focus) value)) nil)))))


(defmethod TRY-DESIGN-OPERATOR (appl (method design-method)
                                     (design-op multiple-design-extension-operator)
                                     (state design-state)  
                                     focus
                                     context)
  (let ((value (apply-design-operator 
                design-op state  focus context)))
  (cond ((eq value :nothing)
         (backtrack-from-operator appl method state design-op focus context))
        (t
         (setf (nogood-values state)
               (cons value (nogood-values state)))
         (when *trace-design-process?*
           (print-with-spaces 2 ;;; (state-depth state)
                              "New assignment is <~a ~A>" (name focus) value))
         (add-design-state appl  method state  (List (cons (name focus) value)) nil)))))
                             

(defmethod BACKTRACK-FROM-OPERATOR (appl 
                                    (method design-method) 
                                    (state design-state) 
                                    (design-op design-operator)
                                    focus
                                    context)
  (when *trace-design-process?*
    (print-with-spaces 2 ;;; (state-depth state)
                       "Operator ~a fails..trying another one..." (name design-op)))
  (design-from-focus appl method state focus context))

(defmethod  APPLY-DESIGN-OPERATOR ((design-op  basic-design-extension-operator)
                                   (state design-state)  
                                   (focus parameter)
                                   context)
  context ;;;ignore
  (with-slots (applicability-env) design-op
    (ocml-apply (the-slot-value (name design-op) 'body)
                (List (name focus))
                applicability-env)))


(defmethod  APPLY-DESIGN-OPERATOR ((design-op  multiple-design-extension-operator)
                                   (state design-state)  
                                   (focus parameter)
                                   context)
  context ;;;ignore
  (with-slots (applicability-env) design-op
    (ocml-apply (the-slot-value (name design-op) 'body)
                (List (name focus)
                      (list 'quote (nogood-values state)))
                applicability-env)))





;;;;;;;;;;;;;;;;;;;;;

(defmethod EVALUATE-STATE-COST (appl (method design-method)(state design-state)
                                     )
  appl
  (with-slots (design-model) state
      (ocml-call *cost-function* `',design-model)))

(defmethod cheaper-state? ((state1 design-state)(state2 design-state))
  (let ((c1 (state-cost state1))
        (c2 (state-cost state2)))
    (cheaper-cost? c1 c2)))

(defun cheaper-cost? (c1 c2)
  (goal-holds? `(,*cost-order-rel* ',c1 ',c2)))

(defun improvement-margin (c1 c2)
  (ocml-call *cost-difference-fun* c1 c2))

(defun add-costs (c1 c2)
  (ocml-call *cost-sum-fun* `',c1 `',c2))



                    


  

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun extend-evaluate-select-control (appl method state)
;; propose&backtrack-control
  (if (state-feasible? state)
    (if (violated-constraints state)
      (state-fails state)
      (if (state-complete? state)
        (design-succeeds state)
        (extend-design appl method state)))
    (state-fails state)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(def-class PROPOSE&BACKTRACK (design-method)
   :lisp-class-name propose&backtrack)

(defmethod DESIGN-FROM-STATE (appl (method propose&backtrack) (state design-state))
  (extend-evaluate-select-control appl method state))

  ;;;;;(propose&backtrack-control appl method state))




  
       
        