;;; -*- Mode: LISP; Syntax: Common-lisp; Base: 10; Package: OCML;   -*-

(in-package "OCML")


(defvar *design-procedures*)
(defvar *design-fixes*)

(defmethod TOP-LEVEL-DESIGN (appl (method propose&revise))
  (setf  *design-procedures* 
         (mapcar #'(lambda (x)
                     (find-current-instance x 'design-procedure))
                 (the-slot-value (name method) 'has-design-procedures))
         *design-fixes* 
         (mapcar #'(lambda (x)
                     (find-current-instance x 'design-fix))
                 (the-slot-value (name method) 'has-design-fixes)))
  
  (cache-design-specification appl method)
  (clear-cached-links2)
  (cache-param-to-ops-links2)
  (cache-param-to-constraints-links)
  (cache-constraints-to-fixes-links)
  (construct-dependency-network)
  (unless (boundp '*dp-breaks*)
    (setf *dp-breaks* (breaks-in-dependency-network)))
  (initialise-design-space appl method)
  (top-level-design-loop appl method))


(defun clear-cached-links2 ()
  (loop for op in *design-procedures*
        do
        (setf ( effective-parameter-list op) nil))
  
  (loop for op in *design-fixes*
        do
        (setf (effective-constraint-list op) nil))
  
  (loop for p in *design-parameters*
        do
        (setf (relevant-operators p) nil)
        (setf (relevant-constraints p) nil))
  
   (loop for c in *design-constraints*
        do
        (setf (effective-parameter-list c) nil)
        (when (instance-of? (name c)
                            'fixable-parametric-constraint)
          (setf (relevant-operators c) nil))))


(defun cache-param-to-ops-links2 ()
  (loop for op in *design-procedures*
        for params = (compute-effective-applicability-set op)
        do
        (loop for param in params
              do
              (link-param-to-operator (find-current-instance param 'parameter)
                                      op))))

(defun cache-constraints-to-fixes-links ()
  (loop for op in *design-fixes*
        for constraints = (compute-effective-applicability-set op)
        do
        (loop for c in constraints
              do
              (link-constraint-to-fix (find-current-instance c 'constraint)
                                      op))))

(defmethod link-constraint-to-fix ((c constraint) (op design-fix))
    (with-slots (effective-constraint-list) op
      (pushnew c effective-constraint-list)
      (pushnew op  (relevant-operators c))))

(defmethod  compute-effective-applicability-set ((d design-fix))
  (the-slot-value (name d) 'applicable-to-constraints))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod DESIGN-FROM-STATE (appl (method propose&revise) (state design-state))
  
  (if (state-feasible? state)
    (cond( (violated-constraints state)
           (revise-design appl method state))
         (t
          ;;;;;(setf *trace-design-process?* nil)
          (if (state-complete? state)
            (design-succeeds state)
            (Progn
              (setf *design-space* (remove state *design-space*))
              
              (extend-design appl method state)))))
    (state-fails state)))

(defmethod SELECT-DESIGN-STATE (appl (method propose&revise) space)
  
  appl
  (when space
    (let ((state (car space)))
      (when (violated-constraints state)
        (setf state (car (sort  
                          (minimal-set 
                           (maximal-set space #'(lambda (state)
                                                  (length (design-model state))))
                           #'(lambda (state)
                               (length (violated-constraints state))))
                          #'(lambda (x y)
                              (cheaper-vector-cost? (state-cost x)(state-cost y)))))))
      (if *trace-design-process?*
        (when state
          (lines 2)
          (print-with-spaces 2 ;;; 2 ;;;  (state-depth (car space)) 
                             "Current state is now ~s"
                             state))
        (when (and state (violated-constraints state))
          (lines 2)
          (print-with-spaces 2 ;;; 2 ;;;  (state-depth (car space)) 
                             "Current state is now ~s"
                             state)))
      state)))


(defmethod REVISE-DESIGN (appl (method design-method) (state design-state))
 ;;(break)
  (let ((*trace-design-process?* t))
  (cond ((new-state? state)
         (setf (state-foci state) (collect-fixable-constraints appl method state)
               (new-state? state) nil)
         (if (state-foci state)
           (design-from-context appl method state :revise)
           (state-fails state)))
        (t
         (backtrack-from-operator appl method state (current-operator state)
                                  (current-focus state) :revise)))))


(defmethod COLLECT-FIXABLE-CONSTRAINTS (appl (method design-method) (state design-state))
  appl
  (with-slots (violated-constraints) state
    (let ((instances (all-instances 'fixable-parametric-constraint)))
      (filter violated-constraints 
              #'(lambda (x)
                  (and (member x instances) 
                       (relevant-operators x)))))))

;;;APPLY-DESIGN-OPERATOR - basic-design-fix - fixable-parametric-constraint
;;;The assignment which violates the constraint is retrieved and the 
;;;fix applied to the culpright parameter, its value, and current design model
(defmethod  APPLY-DESIGN-OPERATOR ((design-op basic-design-fix)
                                   (state design-state)  
                                   (focus fixable-parametric-constraint)
                                   (context (eql :revise)))
(let ((ass (right-value focus (bad-assignments state))))
     (apply-design-operator design-op state 
                            (find-current-instance (car ass) 'parameter)
                            context)))



;;;;;;;;;;;;;;;;;;;;
;;;CMR definitions


 
(defmethod DESIGN-FROM-STATE (appl (method cmr) (state design-state))
  ; (print state)
  (if (state-feasible? state)
    (if (state-complete? state)
      (if (violated-constraints state)
        (revise-design appl method state)
        (design-succeeds state))
      (Progn
        (setf *design-space* (remove state *design-space*))
       
        (extend-design appl method state)))
    (state-fails state)))

(defmethod SELECT-DESIGN-STATE (appl (method cmr) space)
  appl
  (let ((state (car space)))
    (when state
      (if (state-complete? state)
        (car (sort (minimal-set space #'(lambda (state)
                               (length (violated-constraints state))))
                   #'(lambda (x y)
                       (cheaper-vector-cost? (state-cost x)(state-cost y)))))
        state))))

(defmethod EVALUATE-STATE-INTERNAL (appl (method cmr)(state design-state))
   (setf (state-feasible? state) (evaluate-feasibility appl method state)  
        (state-complete? state) (design-complete? appl method state)
        (violated-constraints state) (when (state-complete? state)
                                       (check-dm-constraints (design-model state)))))