;;; -*- Mode: LISP; Syntax: Common-lisp; Base: 10; Package: OCML; -*- 

(in-package "OCML")

(defvar *trace-classifier?* nil)

(defun trace-classifier ()
  (setf *trace-classifier?* t))

(defun untrace-classifier ()
  (setf *trace-classifier?* nil))


(in-ontology simple-classification)


(def-class simple-classification-psm (problem-solving-method)
  "This methods simple iterates over all candidates, classifying
   the instance with respect to each candidate and then 
   returning the best explaining classes"
  ((has-input-role :value candidate-classes
                   :value candidate-instance)
   (has-output-role :value best-matching-classes)
   (candidate-classes :type list)
   (candidate-instance :type instance)
   (best-matching-classes :type list)
   (has-body :value (lambda (?psm)
                      (simple-classification-procedure ?psm))))
  :own-slots ((tackles-task-type  classification-task)
              (has-generic-subtasks '(match-class compare-match-scores)))
  :lisp-class-name simple-classification-psm)


(def-procedure simple-classification-procedure (?psm)
  :lisp-fun #'(lambda (psm)
                (simple-classify psm)))

;;;A simple classification method.  It has to return the 
;;;classes which best explain the data - i.e. the instance.
(defun simple-classify (method)
  (let* ((instance  (get-role-value method 'candidate-instance))
        (candidates  (get-role-value method 'candidate-classes))
        (features (setofall '?x `(has-feature ,instance ?x))))
    (when candidates
      (loop with scores = (cons (achieve-generic-subtask
                                 method 'match-class 
                                 'candidate-instance instance
                                 'has-features features
                                 'candidate-class (car candidates))
                                (List (car candidates)))
            for c in (cdr candidates)
            for score = (achieve-generic-subtask
                         method 'match-class 
                         'candidate-instance instance
                         'has-features features
                         'candidate-class c)
            for best-score = (achieve-generic-subtask
                              method 'compare-match-scores
                              'has-score1 score
                              'has-score2 (car scores))
            do
            (cond ((equal best-score :equal)
                   (push c (cdr scores)))
                  ((equal best-score score)
                   (setf scores (cons score
                                 (list c)))))
            finally
            (return (cdr scores))))))

;;;MATCH-CLASS
(def-task match-class (goal-specification-task) ?task 
  ((has-input-role :value candidate-class
                   :value has-features
                   :value candidate-instance)
   (has-output-role :value has-score)
   (candidate-class :type class)
   (candidate-instance :type instance)
   (has-features :type list)
   (has-score :type match-score)
   (has-goal-expression 
    :value (kappa (?task ?score)
                  (and (match-score ?score)
                       (has-score ?task ?score))))))


;;; BASIC-MATCH-CLASS-PSM 
(def-class basic-match-class-psm (problem-solving-method) ?psm
  "This psm uses the standard-class-instance-match-score defined
   in the classification task ontology"
  ((has-body :value (lambda (?psm)
                      (simple-match-class ?psm)))
   (has-postcondition (kappa (?psm ?score)
                             (standard-class-instance-match-score 
                              ?score))))
  :own-slots ((tackles-task-type  match-class)
              (has-generic-subtasks '(match-feature)))
  :lisp-class-name basic-match-class-psm)

(def-procedure simple-match-class (?psm)
  :lisp-fun #'(lambda (psm)
                (simple-match-class  psm)))

;;;SIMPLE-MATCH-CLASS
;;;Here we use a simple model of class matching.  GIven a feature f1 in instance I 
;;;and a class C, there can be 3 cases

;;;Feature is INCONSISTENT - f1 is in I and in C but their values are inconsistent
;;;Feature is EXPLAINED - f1 is in I and in C and their values are consistent
;;;Feature is UNEXPLAINED - f1 is in I but not in C

;;;It is also possible that teh class C has a feature fc, which is not in in I. 
;;;In this case we say that the feature is MISSING - i.e.

;;;Feature is MISSING - f1 is in C but not in I

;;;Hence, the result of matching a class to an instance is a 4-place vector
;;;<inconsistent-features, unexplained-features, explained-features, missing-features>

(defun simple-match-class (method)
  (let* ((instance  (get-role-value method  'candidate-instance))
        (class  (get-role-value method  'candidate-class))
        (instance-features  (get-role-value method  'has-features))
        (class-features (setofall '?x `(has-feature ,class  ?x)))
        (score-vector (list nil nil nil nil)))

    (when *trace-classifier?*
      (lines 2)
      (print-with-spaces 2 
                         "Matching class ~S...."class))


    ;;initialize unexplained features (in I but not in C)
    (setf (third score-vector)     
          (set-difference instance-features 
                          class-features))
    (loop for feature in instance-features
          for result = (achieve-generic-subtask
                        method 'match-feature 
                        'candidate-instance instance
                        'class-features class-features
                        'candidate-feature feature
                        'candidate-class class)
          do
          (case result
            (:inconsistent (setf (first score-vector)
                                 (cons feature (first score-vector))))
            (:explained (setf (second score-vector)
                                 (cons feature (second score-vector))))
            (:unexplained (setf (third score-vector)
                                 (cons feature (third score-vector)))))
          finally
          (progn 
            (setf (fourth score-vector)
                (set-difference class-features instance-features))
            (when *trace-classifier?*
              (lines 2)
              (print-with-spaces 2 
                                 "Result is ~s "score-vector))
            (return score-vector)))))


;;;MATCH-FEATURE
(def-task match-feature (goal-specification-task) ?task 
  ((has-input-role :value candidate-class
                   :value class-features
                   :value candidate-feature
                   :value candidate-instance)
   (has-output-role :value has-feature-score)
   (candidate-class :type class)
   (candidate-instance :type instance)
   (candidate-feature :type list)
   (class-features :type list)
   (has-feature-score :type feature-score)
   (has-goal-expression 
    :value (kappa (?task ?score)
                  (and (feature-score ?score)
                       (has-feature-score ?task ?score))))))

;;;BASIC-MATCH-FEATURE-PSM
(def-class basic-match-feature-psm (problem-solving-method)
  "A psm for task match-feature that subscribes to the 
   typology of feature-match-score defined by class
   standard-feature-score"
   
  ((has-body :value (lambda (?psm)
                      (simple-match-class-feature ?psm)))
   (has-postcondition (kappa (?psm ?score)
                             (standard-feature-score ?score))))
  :own-slots ((tackles-task-type  match-feature))
  :lisp-class-name basic-match-feature-psm)

(def-procedure simple-match-class-feature (?psm)
  :lisp-fun #'(lambda (psm)
                (simple-match-class-feature psm)))

;;;Here we assume that an instance can have only one value for the
;;;feature in question, while a class can have multiple ones with
;;;an implicit disjunction - i.e an instance has to match only one 
;;;of these.
(defun simple-match-class-feature (method)
  (let* ((instance  (get-role-value method  'candidate-instance))
         (class  (get-role-value method  'candidate-class))
         (feature  (get-role-value method  'candidate-feature))
         (class-features (get-role-value method  'class-features))
         (instance-value (ocml-eval-gen `(feature-value ,instance ,feature))))
    (cond ((member feature class-features)
           (let ((class-values (all-class-slot-values class feature)))
             (if class-values
               (if (some #'(lambda (value)
                             (not (eq :fail
                                      (unify instance-value
                                             value))))
                       class-values)
                 :explained
                 :inconsistent)
               :unexplained)))
          (t
           :unexplained))))

(def-task compare-match-scores (goal-specification-task) ?task 
  ((has-input-role :value has-score1 
                   :value has-score2)
   (has-output-role :value has-best-score)
   (has-score1 :type match-score)
   (has-score2 :type match-score)
   (has-best-score :type match-score-or-equal)
   (has-goal-expression 
    :value (kappa (?task ?score)
                  (and (match-score-or-equal ?score)
                       (has-best-score ?task ?score))))))

;;;CLASS MATCH-SCORE-OR-EQUAL
(def-class match-score-or-equal () ?x
  :iff-def (or (match-score ?x)
               (= ?x :equal)))


;;;SIMPLE-COMPARE-MATCH-SCORE-PSM
(def-class simple-compare-match-score-psm (problem-solving-method) ?psm
  "A match-score comparison psm which relies on scores 
   adhering to the standard-class-instance-match-score format"
  ((has-body :value (lambda (?psm)
                      (simple-compare-match-scores ?psm))))
  :constraint (and (standard-class-instance-match-score 
                    (role-value ?psm has-score1))
                   (standard-class-instance-match-score 
                    (role-value ?psm has-score2)))
  :own-slots ((tackles-task-type compare-match-scores))
  :lisp-class-name simple-compare-match-score-psm)


(def-procedure simple-compare-match-scores (?psm)
  :body (the-better-match-score 
         (role-value ?psm 'has-score1)
         (role-value ?psm 'has-score2)))


;(defun simple-compare-match-scores (method)
;  (let ((score1  (get-role-value method 'has-score1 ))
;        (score2  (get-role-value method 'has-score2 )))
;    (ocml-eval-gen `(the-better-match-score ,score1 ,score2))))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun classify (&key top-class 
                      feature-value-pairs
                      (method-type 'simple-classification-psm)
                      (domain-name top-class)
                               )
  (let* ((instance (define-domain-instance 
                    (gentemp (string top-class))
                    'classificandum
                    ""
                    `((has-feature-value-pairs ,feature-value-pairs))))
         (candidates (setofall '?x `(subclass-of ?x ,top-class)))
         (task (define-domain-instance 
                 (gentemp (string-append (string top-class) "-"
                                                 "CLASSIFICATION-TASK"))
                 'classification-task
                ""
                `((candidate-classes ,candidates)
                  (candidate-instance ,(name instance)))))
         (method (define-domain-instance 
                   (gentemp "METHOD")
                   method-type
                   ""
                   ()))
         (application (define-domain-instance 
                        (gentemp (string-append (string top-class) 
                                                "-"
                                                 "CLASSIFICATION"))
                        'application
                        ""
                        `((tackles-domain ,domain-name)
                          (uses-method ,(name method))
                          (tackles-task ,(name task))))))
    (ocml-eval-gen `(solve-application ,(name application)))))
               
               
                
                   


  



  






